from yta_general_utils.programming.enum import YTAEnum as Enum


class MANIM_RENDERER(Enum):
    """
    The different rendeders manim has available.

    These are some tests I made with these renderers:
    - ImageMobject + Cairo works, but positioning gets crazy.
    - ImageMobject + Opengl fails
    - OpenGLImageMobject + Opengl works perfectly.
    - VideoMobject (ImageMobject) + Cairo works, but positioning gets crazy.
    - VideoMobject (ImageMobject) + Opengl fails
    - VideoMobject (OpenGLImageMobject) + Opengl only shows the first frame, but positioning is perfect.
    - Didn't test anything else
    """
    CAIRO = 'cairo'
    OPENGL = 'opengl'


from yta_multimedia.video.generation.manim.classes.base_manim_animation import BaseManimAnimation, NewBaseManimAnimation
from yta_multimedia.video.generation.manim.enums import MANIM_RENDERER
from yta_general_utils.programming.parameter_validator import NumberValidator, PythonValidator
from yta_general_utils.programming.attribute_obtainer import AttributeObtainer
from manim import *
from typing import Union


# TODO: This base class should be next to
# BaseManimAnimation class
class BaseManimAnimationWrapper:
    """
    Base class for all the mani animation generator
    classes that we want to have in our system.

    This wrapper is to define the attributes that 
    are needed and the manim animation generator
    class that will be used to generate it.

    TODO: maybe we can simplify this to a unique
    class that is not only the wrapper but also the
    generator, as the generator classes will be 
    refactored with this new format.
    """
    animation_generator_instance: NewBaseManimAnimation = None

    def __init__(self, animation_generator: NewBaseManimAnimation):
        if not PythonValidator.is_subclass(animation_generator, NewBaseManimAnimation):
            raise Exception('The provided "animation_generator" is not a subclass nor an instance of a subclass of NewBaseManimAnimation class.')
        
        if PythonValidator.is_a_class(animation_generator):
            animation_generator = animation_generator()

        self.animation_generator_instance = animation_generator

    @property
    def attributes(self):
        """
        Only the values that are actually set on the
        instance are obtained with 'vars'. If you set
        'var_name = None' but you don't do 
        'self.var_name = 33' in the '__init__' method,
        it won't be returned by the 'vars()' method.
        """
        # The variables we don't want have to be
        # manually set if we make any change
        return AttributeObtainer.get_attributes_from_instance(self, ['animation_generator_instance', 'attributes'])

    def build(self):
        """
        Build the manim animation if the parameters are
        valid and returns the filename of the generated
        video to be used in the app (you should handle it
        with a 'VideoFileClip(o, has_mask = True)' to load
        it with mask and to be able to handle it).
        """
        return self.animation_generator_instance.generate(self.attributes, output_filename = 'output.mov')
    
# This class and the generator must be always in
# the same file because they are extremely
# connected and they are dependant
class ExampleManimAnimationWrapper(BaseManimAnimationWrapper):
    """
    Example manim wrapper class that collects the 
    parameters, validates them and builds the scene
    using the associated generator class.
    """
    text: str = None
    duration: float = None

    def __init__(self, text: str, duration: float):
        # TODO: This parameter validation could be done
        # by our specific parameter validator (under
        # construction yet)
        exception_messages = []

        if not text:
            exception_messages.append('No "text" parameter provided.')
        
        if not NumberValidator.is_positive_number(duration):
            exception_messages.append('The "duration" parameter provided is not a positive number.')

        if len(exception_messages) > 0:
            raise Exception('\n'.join([exception_message for exception_message in exception_messages]))
        
        self.text = text
        self.duration = duration
        super().__init__(ExampleManimAnimationGenerator)

class ExampleManimAnimationGenerator(NewBaseManimAnimation):
    def construct(self):
        """
        This method is called by manim when executed by shell and
        will call the scene animation render method to be processed
        and generated.
        """
        self.animate()

    def generate(self, parameters: dict, output_filename: Union[str, None] = None):
        """
        Build the manim animation video and stores it
        locally as the provided 'output_filename', 
        returning the final output filename.

        This method will call the '.animate()' method
        to build the scene with the instructions that
        are written in that method.
        """
        return super().generate(
            parameters,
            renderer = MANIM_RENDERER.CAIRO,
            output_filename = output_filename
        )
    
    def animate(self):
        text = Text(self.parameters['text'], font_size = 26, font = 'Minecraftia').shift(DOWN * 0).scale(1)
        self.add(text)
        # The sound duration will be set as video duration if
        # larger than the 'duration' parameter so I need to
        # look for a way to enshort the video
        #self.add_sound(GoogleDriveResource('https://drive.google.com/file/d/1Dzeb6Qae4UdpmuA6U9d6t3MXnvhzukzg/view?usp=sharing').download(create_temp_filename('tmp.mp3')))
        self.wait(self.parameters['duration'])













# TODO: Remove this below when the 2 'example' classes
# above are working perfectly with the new base class
class SimpleTextManimAnimationX:
    text: str = None
    duration: float = None
    _animation: callable = None

    def __init__(self, text: str, duration: float):
        exception_messages = []

        if not text:
            exception_messages.append('No "text" parameter provided.')
        
        if not NumberValidator.is_positive_number(duration):
            exception_messages.append('The "duration" parameter provided is not a positive number.')

        if len(exception_messages) > 0:
            raise Exception('\n'.join([exception_message for exception_message in exception_messages]))
        
        self.text = text
        self.duration = duration
        self._animation = SimpleTextManimAnimation

    @property
    def attributes(self):
        """
        Only the values that are actually set on the
        instance are obtained with 'vars'. If you set
        'var_name = None' but you don't do 
        'self.var_name = 33' in the '__init__' method,
        it won't be returned by the 'vars()' method.
        """
        # The variables we don't want have to be
        # manually set if we make any change
        # Each attribute is a dict 'key: value'
        return {k: v for k, v in vars(self).items() if k not in ['_animation', 'attributes']}
    
    def generate(self):
        """
        Generate the manim animation if the parameters are
        valid and returns the filename of the generated
        video to be used in the app (you should handle it
        with a 'VideoFileClip(o, has_mask = True)' to load
        it with mask and to be able to handle it).
        """
        return self._animation().generate(**self.attributes)

# class SimpleTextManimAnimation(BaseManimAnimation):
#     # TODO: Maybe make this more strict with 'type' also
#     required_parameters = {
#         'text': MANDATORY_CONFIG_PARAMETER,
#         'duration': MANDATORY_CONFIG_PARAMETER,
#     }

#     def construct(self):
#         """
#         This method is called by manim when executed by shell and
#         will call the scene animation render method to be processed
#         and generated.
#         """
#         self.animate()

#     def generate(self, text: str, duration: float, output_filename: str = 'output.mov'):
#         """
#         Checks and validates the provided parameters and generates
#         the manim animation if those parameters are valid.
#         """
#         # Check and validate all parameters
#         parameters = {}

#         if super().parameter_is_mandatory('text', self.required_parameters) and not text:
#             raise Exception('Field "text" is mandatory. Aborting manim creation...')
        
#         parameters['text'] = text

#         if super().parameter_is_mandatory('duration', self.required_parameters) and not duration:
#             raise Exception('Field "duration" is mandatory. Aborting manim creation...')
#         if duration < 0 or duration > 100:
#             raise Exception('Field "duration" value is not valid. Must be between 0 and 100')
        
#         parameters['duration'] = duration

#         if not output_filename:
#             output_filename = 'output.mov'

#         # Generate the animation when parameters are valid
#         super().generate(parameters, output_filename = output_filename)

#         return output_filename
    
#     def animate_one(self):
#         """
#         This code will generate the manim animation and belongs to the
#         Scene manim object.
#         """
#         text = Text(self.parameters['text'], font_size = 140, stroke_width = 2.0, font = 'Haettenschweiler').shift(DOWN * 0).scale(0.001)
#         self.wait(1 / 60)
#         self.add(text)
#         self.add_sound(GoogleDriveResource('https://drive.google.com/file/d/1WPS8uWB1LTuzPzxQ2Zcp1FpwvLM3fhM5/view?usp=sharing').download(create_temp_filename('tmp.mp3')))
#         self.play(text.animate.scale(1000), run_time = 49 / 60)
#         self.play(Rotate(text, 0.03), run_time = 3 / 60)
#         self.play(Rotate(text, -0.04), run_time = 3 / 60)
#         self.play(Rotate(text, -0.02), run_time = 3 / 60)
#         self.play(Rotate(text, 0.04), run_time = 3 / 60)
#         self.play(Rotate(text, -0.01), run_time = 3 / 60)
#         self.play(Rotate(text, 0.03), run_time = 3 / 60)
#         self.play(Rotate(text, -0.04), run_time = 3 / 60)
#         self.play(Rotate(text, -0.02), run_time = 3 / 60)
#         self.play(Rotate(text, 0.04), run_time = 3 / 60)
#         self.play(Rotate(text, -0.01), run_time = 3 / 60)
#         #self.add_sound(TOUSE_ABSOLUTE_PATH + 'sounds/xp_error.mp3')
#         #self.play(AddTextLetterByLetter(text), run_time = self.parameters['duration'])
        
#         #simple_play_animation(self, Write, text, self.parameters['duration'])

#     def animate_two(self):
#         """
#         This code will generate the manim animation and belongs to the
#         Scene manim object.
#         """
#         text = Text(self.parameters['text'], font_size = 140, stroke_width = 2.0, font = 'Haettenschweiler').shift(DOWN * 0).scale(7 / 10)
#         self.wait(1 / 60)
#         self.add(text)
#         self.play(text.animate.scale(10 / 7), run_time = 6 / 60)
#         self.play(ApplyWave(text), run_time = 30 / 60)

#     def animate(self):
#         text = Text(self.parameters['text'], font_size = 26, font = 'Minecraftia').shift(DOWN * 0).scale(1)
#         self.add(text)
#         # The sound duration will be set as video duration if
#         # larger than the 'duration' parameter so I need to
#         # look for a way to enshort the video
#         #self.add_sound(GoogleDriveResource('https://drive.google.com/file/d/1Dzeb6Qae4UdpmuA6U9d6t3MXnvhzukzg/view?usp=sharing').download(create_temp_filename('tmp.mp3')))
#         self.wait(self.parameters['duration'])

# __all__ = [
#     'SimpleTextManimAnimationX'
# ]


from yta_general_utils.temp import create_temp_filename
from yta_general_utils.image.parser import ImageParser
from PIL import Image


def pixelate_image(image, i_size, output_filename: str):
    """
    Pixelates the provided 'image' and saves it as the 'output_filename' if
    provided.
    The 'i_size' is the pixelating square. The smaller it is, the less pixelated 
    its.

    'i_size' must be a tuple such as (8, 8) or (16, 16).
    """
    img = ImageParser.to_pillow(image)

    # Convert to small image
    small_img = img.resize(i_size, Image.BILINEAR)

    # Resize to output size
    res = small_img.resize(img.size, Image.NEAREST)

    # TODO: Validate 'output_filename' if provided
    if not output_filename:
        output_filename = create_temp_filename('tmp_sketch.png')
        
    # TODO: Return it instead of saving
    res.save(output_filename)


class DiscordConversationCharacter:
    def __init__(self, username: str, avatar_url: str):
        if not username:
            raise Exception('No "username" provided.')
        
        if not avatar_url:
            raise Exception('No "avatar_url" provided.')
        
        # TODO: Check if 'avatar_url' is valid

        # TODO: Maybe if they don't provide this information we
        # could auto-generate it
        self.username = username
        self.avatar_url = avatar_url


"""
Some tests I made with Manim (3D, as it is the most difficult):
- ImageMobject + Cairo works, but positioning gets crazy.
- ImageMobject + Opengl fails
- OpenGLImageMobject + Opengl works perfectly.
- VideoMobject (ImageMobject) + Cairo works, but positioning gets crazy.
- VideoMobject (ImageMobject) + Opengl fails
- VideoMobject (OpenGLImageMobject) + Opengl only shows the first frame, but positioning is perfect.
- Didn't test anything else
"""
from yta_multimedia.video.generation.manim.classes.base_manim_animation import BaseManimAnimation
from yta_multimedia.video.generation.manim.classes.base_three_d_manim_animation import BaseThreeDManimAnimation
from yta_multimedia.video.generation.manim.classes.video.mobjects.video_mobject import VideoMobject
from yta_multimedia.video.generation.manim.classes.video.mobjects.video_opengl_mobject import VideoOpenGLMobject
from yta_multimedia.video.generation.manim.classes.image.opengl_image_mobject import OpenGLImageMobject
from manim import *


class TestVideoMobjectIn2DManimAnimation(BaseManimAnimation):
    def construct(self):
        """
        This method is called by manim when executed by shell and
        will call the scene animation render method to be processed
        and generated.
        """
        self.animate()

    def generate(self, output_filename: str = 'output.mov'):
        """
        Checks and validates the provided parameters and generates
        the manim animation if those parameters are valid. The 
        'text' parameter is limited to 30 characters.
        """
        # Check and validate all parameters
        parameters = {}

        # Generate the animation when parameters are valid
        super().generate(parameters, renderer = 'cairo', output_filename = output_filename)

        return output_filename
    
    def animate(self):
        video1 = VideoMobject(
            filename = 'prueba.mp4',
        ).scale_to_fit_width(5)
        self.add(video1)
        self.wait(0.25)
        self.play(video1.animate.shift(1 * UP), run_time = 0.25)
        self.play(video1.animate.shift(2 * DOWN), run_time = 0.5)
        self.play(video1.animate.shift(1 * UP), run_time = 0.25)
        self.wait(0.25)

class TestVideoOpenGLMobjectIn2DManimAnimation(BaseManimAnimation):
    def construct(self):
        """
        This method is called by manim when executed by shell and
        will call the scene animation render method to be processed
        and generated.
        """
        self.animate()

    def generate(self, output_filename: str = 'output.mov'):
        """
        Checks and validates the provided parameters and generates
        the manim animation if those parameters are valid. The 
        'text' parameter is limited to 30 characters.
        """
        # Check and validate all parameters
        parameters = {}

        # Generate the animation when parameters are valid
        super().generate(parameters, renderer = 'opengl', output_filename = output_filename)

        return output_filename
    
    def animate(self):
        video1 = VideoOpenGLMobject(
            filename = 'prueba.mp4',
        ).scale_to_fit_width(5)
        self.add(video1)
        self.wait(0.25)
        self.play(video1.animate.shift(1 * UP), run_time = 0.25)
        self.play(video1.animate.shift(2 * DOWN), run_time = 0.5)
        self.play(video1.animate.shift(1 * UP), run_time = 0.25)
        self.wait(0.25)

class TestVideoMobjectIn3DManimAnimation(BaseThreeDManimAnimation):
    def construct(self):
        """
        This method is called by manim when executed by shell and
        will call the scene animation render method to be processed
        and generated.
        """
        self.animate()

    def generate(self, output_filename: str = 'output.mov'):
        """
        Checks and validates the provided parameters and generates
        the manim animation if those parameters are valid. The 
        'text' parameter is limited to 30 characters.
        """
        # Check and validate all parameters
        parameters = {}

        # Generate the animation when parameters are valid
        super().generate(parameters, renderer = 'cairo', output_filename = output_filename)

        return output_filename
    
    def animate(self):
        video1 = VideoMobject(
            filename = 'prueba.mp4',
        ).scale_to_fit_width(5)
        self.add(video1)
        self.wait(0.25)
        self.begin_ambient_camera_rotation(rate = 0.15)
        self.play(video1.animate.shift(1 * UP), run_time = 0.25)
        self.play(video1.animate.shift(2 * DOWN), run_time = 0.5)
        self.play(video1.animate.shift(1 * UP), run_time = 0.25)
        self.move_camera(phi=75 * DEGREES, theta=30 * DEGREES, zoom=1, run_time=1.5)
        self.wait(0.25)

class TestVideoOpenGLMobjectIn3DManimAnimation(BaseThreeDManimAnimation):
    def construct(self):
        """
        This method is called by manim when executed by shell and
        will call the scene animation render method to be processed
        and generated.
        """
        self.animate()

    def generate(self, output_filename: str = 'output.mov'):
        """
        Checks and validates the provided parameters and generates
        the manim animation if those parameters are valid. The 
        'text' parameter is limited to 30 characters.
        """
        # Check and validate all parameters
        parameters = {}

        # Generate the animation when parameters are valid
        super().generate(parameters, renderer = 'opengl', output_filename = output_filename)

        return output_filename
    
    def animate(self):
        video1 = VideoOpenGLMobject(
            filename = 'prueba.mp4',
        ).scale_to_fit_width(5)
        self.add(video1)
        self.wait(0.25)
        self.begin_ambient_camera_rotation(rate = 0.15)
        self.play(video1.animate.shift(1 * UP), run_time = 0.25)
        self.play(video1.animate.shift(2 * DOWN), run_time = 0.5)
        self.play(video1.animate.shift(1 * UP), run_time = 0.25)
        self.move_camera(phi=75 * DEGREES, theta=30 * DEGREES, zoom=1, run_time=1.5)
        self.wait(0.25)

class TestImageOpenGLMobjectIn3DManimAnimation(BaseThreeDManimAnimation):
    def construct(self):
        """
        This method is called by manim when executed by shell and
        will call the scene animation render method to be processed
        and generated.
        """
        self.animate()

    def generate(self, filename: str, output_filename: str = 'output.mov'):
        """
        Checks and validates the provided parameters and generates
        the manim animation if those parameters are valid. The 
        'text' parameter is limited to 30 characters.
        """
        # Check and validate all parameters
        parameters = {
            'filename': filename
        }

        # Generate the animation when parameters are valid
        super().generate(parameters, renderer = 'opengl', output_filename = output_filename)

        return output_filename
    
    def animate(self):
        image = OpenGLImageMobject.init(self.parameters['filename']).scale_to_fit_width(5)
        self.add(image)
        image.shift(10 * UP)
        self.play(image.animate.shift(10 * DOWN), run_time = 0.25)
        self.move_camera(phi = 30 * DEGREES, frame_center = image, run_time = 0.25)
        self.move_camera(phi = -60 * DEGREES, frame_center = image, run_time = 0.50)
        self.move_camera(phi = 30 * DEGREES, frame_center = image, run_time = 0.25)
        self.wait(0.25)

class TestOpenGLImageMobjectIn3DManimAnimation(BaseThreeDManimAnimation):
    def construct(self):
        """
        This method is called by manim when executed by shell and
        will call the scene animation render method to be processed
        and generated.
        """
        self.animate()

    def generate(self, output_filename: str = 'output.mov'):
        """
        Checks and validates the provided parameters and generates
        the manim animation if those parameters are valid. The 
        'text' parameter is limited to 30 characters.
        """
        # Check and validate all parameters
        parameters = {}

        # Generate the animation when parameters are valid
        super().generate(parameters, renderer = 'opengl', output_filename = output_filename)

        return output_filename
    
    def animate(self):
        from yta_multimedia.image.generation.reviews.tripadvisor.tripadvisor_image_generator import TripadvisorImageGenerator
        from yta_multimedia.video.consts import MANIM_SCENE_DEFAULT_SIZE

        # This makes an image navigating in 3D scenario and being rendered
        # with alpha transparency (if it nos working, go to 'opengl_renderer.py'
        # and change the line 596 replacing 1.0 by 0.0)
        self.camera.background_color = [0, 0, 0, 0]
        image = OpenGLImageMobject.init(
            filename_or_array = np.asarray(TripadvisorImageGenerator.generate_review()),
        ).scale_to_fit_width(5)
        image.move_to((0, MANIM_SCENE_DEFAULT_SIZE[1] + image.height / 2, 0))
        self.add(image)
        #self.move_camera(phi = 75 * DEGREES, theta = 30 * DEGREES, run_time = 0.0000001)
        self.set_camera_orientation(phi = 75 * DEGREES, theta = 30 * DEGREES)
        self.play(image.animate.move_to((0, -MANIM_SCENE_DEFAULT_SIZE[1] - image.height / 2, 0)), run_time = 5, rate_func = linear)
        print('Animation finished')
        print(self.camera.background_opacity)

    def animate_two(self):
        from yta_multimedia.video.frames.video_frame_extractor import VideoFrameExtractor
        from yta_multimedia.video.consts import MANIM_SCENE_DEFAULT_SIZE
        from moviepy import VideoFileClip

        # This below is for trying to render a video but the internal
        # counter is not working properly
        image = OpenGLImageMobject.init(
            filename_or_array = VideoFrameExtractor.get_frame_by_number('prueba.mp4', 0)
        ).scale_to_fit_width(MANIM_SCENE_DEFAULT_SIZE[0])
        self.add(image)
        #self.begin_ambient_camera_rotation()
        self.wait(0.25)
        self.wait(0.25)
        self.move_camera(phi = 45 * DEGREES, added_anims = [image.animate.move_to((0, MANIM_SCENE_DEFAULT_SIZE[1] / 2, -MANIM_SCENE_DEFAULT_SIZE[1] / 2))], run_time = 1, rate_func = linear)
        videoclip = VideoFileClip('prueba.mp4')
        #previous_image = image
        print(videoclip.fps * videoclip.duration)
        print((int) (videoclip.fps * videoclip.duration))
        videoclip = videoclip.with_subclip(0, 1)
        for i in range((int) (videoclip.fps * videoclip.duration)):
            if i == 0:
                i += 1
                print(self.mobjects)
                other_image = OpenGLImageMobject.init(
                    filename_or_array = VideoFrameExtractor.get_frame_by_number('prueba.mp4', i),
                ).scale_to_fit_width(MANIM_SCENE_DEFAULT_SIZE[0]).move_to(self.mobjects[0])
                print(self.mobjects[0].__class__)
                print(self.mobjects[1].__class__)
                print(self.mobjects)
                self.add(other_image)
                print(self.mobjects)
                self.play(Transform(self.mobjects[0], self.mobjects[len(self.mobjects) - 1]), run_time = 1 / 60)
                print(self.mobjects)
                self.remove(self.mobjects[0])
                image = other_image
                #self.add(other_image)
                #previous_image = other_image
                #self.wait(1 / 60)
                print(self.mobjects)
            else:
                print(self.mobjects)
                other_image = OpenGLImageMobject.init(
                    filename_or_array = VideoFrameExtractor.get_frame_by_number('prueba.mp4', i),
                ).scale_to_fit_width(MANIM_SCENE_DEFAULT_SIZE[0]).move_to(self.mobjects[len(self.mobjects) - 1])
                print(self.mobjects[0].__class__)
                print(self.mobjects[1].__class__)
                print(self.mobjects)
                self.add(other_image)
                print(self.mobjects)
                self.play(Transform(self.mobjects[len(self.mobjects) - 2], self.mobjects[len(self.mobjects) - 1]), run_time = 1 / 60)
                print(self.mobjects)
                self.remove(self.mobjects[len(self.mobjects) - 2])
                image = other_image
                #self.add(other_image)
                #previous_image = other_image
                #self.wait(1 / 60)
                print(self.mobjects)
        #self.play(Uncreate(image), run_time = 0.5, rate_func = linear)
        #image.animate.move_to((0, 0, HALF_SCENE_HEIGHT))
        self.wait(0.25)
        #self.stop_ambient_camera_rotation()


from yta_general_utils.downloader import Downloader

import requests


# TODO: Check this because I don't know if this webpage is using the tts (coqui)
# library as the generator engine. If that, I have this engine in 'coqui.py' file
# so I don't need this (that is not stable because is based in http requests)
def narrate_tts3(text: str, output_filename: str):
    """
    This makes a narration based on an external platform. You
    can change some voice configuration in code to make the
    voice different.

    Aparrently not limited. Check, because it has time breaks 
    and that stuff to enhance the narration.
    """
    # From here: https://ttsmp3.com/
    headers = {
        'accept': '*/*',
        'accept-language': 'es-ES,es;q=0.9',
        'content-type': 'application/x-www-form-urlencoded',
        'origin': 'https://ttsmp3.com',
        'referer': 'https://ttsmp3.com/',
        'sec-ch-ua': '"Google Chrome";v="123", "Not:A-Brand";v="8", "Chromium";v="123"',
        'sec-ch-ua-mobile': '?0',
        'sec-ch-ua-platform': '"Windows"',
        'sec-fetch-dest': 'empty',
        'sec-fetch-mode': 'cors',
        'sec-fetch-site': 'same-origin',
        'user-agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/123.0.0.0 Safari/537.36',
    }

    VOICES = ['Lupe', 'Penelope', 'Miguel']
    data = {
        'msg': text,
        'lang': 'Lupe',
        'source': 'ttsmp3',
    }

    response = requests.post('https://ttsmp3.com/makemp3_new.php', headers = headers, data = data)
    response = response.json()
    url = response['URL']
    # "https://ttsmp3.com/created_mp3/8b38a5f2d4664e98c9757eb6db93b914.mp3"
    Downloader.download_audio(url, output_filename)



from yta_multimedia.video.generation.manim.classes.base_manim_animation import BaseManimAnimation
from yta_multimedia.video.consts import MOVIEPY_SCENE_DEFAULT_SIZE
from yta_multimedia.video.generation.manim.utils.dimensions import fitting_text
from yta_multimedia.video.position import Position
from manim import *


class RainOfWordsManimAnimation(BaseManimAnimation):
    """
    This is a rain of the provided 'words' over the screen, that
    appear in random positions.
    """
    def construct(self):
        """
        This method is called by manim when executed by shell and
        will call the scene animation render method to be processed
        and generated.
        """
        self.animate()

    def generate(self, words: list[str], duration: float, output_filename: str = 'output.mov'):
        """
        Checks and validates the provided parameters and generates
        the manim animation if those parameters are valid.
        """
        # Check and validate all parameters
        parameters = {}

        if super().parameter_is_mandatory('words', self.required_parameters) and not words or len(words) <= 0:
            raise Exception('Field "text" is mandatory. Aborting manim creation...')
        
        parameters['words'] = words

        if super().parameter_is_mandatory('duration', self.required_parameters) and not duration:
            raise Exception('Field "duration" is mandatory. Aborting manim creation...')
        if duration < 0 or duration > 100:
            raise Exception('Field "duration" value is not valid. Must be between 0 and 100')
        
        parameters['duration'] = duration

        if not output_filename:
            output_filename = 'output.mov'

        # Generate the animation when parameters are valid
        super().generate(parameters, output_filename = output_filename)

        return output_filename

    def animate(self):
        each_word_time = self.parameters['duration'] / len(self.parameters['words'])
        # Adjust the divisor number to modify word size
        for word in self.parameters['words']:
            text = fitting_text(word, MOVIEPY_SCENE_DEFAULT_SIZE[0] / 6)
            random_coords = Position.RANDOM_INSIDE.get_manim_position((text.width, text.height))
            text.move_to([random_coords['x'], random_coords['y'], 0])
            self.add(text)
            self.wait(each_word_time)
        


from yta_multimedia.audio.parser import AudioParser
from yta_general_utils.programming.parameter_validator import NumberValidator
from pydub import AudioSegment, silence
from typing import Union


class AudioSilence:
    """
    Class to simplify and encapsulate the interaction with audio silences.
    """
    @staticmethod
    def detect(audio, minimum_silence_ms: int = 250):
        """
        Detect the silences of a minimum of 'minimum_silence_ms' milliseconds
        time and returns an array containing tuples with the start and the 
        end of the silence moments.
        """
        audio = AudioParser.as_audiosegment(audio)

        minimum_silence_ms = 250 if not minimum_silence_ms else minimum_silence_ms

        if not NumberValidator.is_positive_number(minimum_silence_ms):
            raise Exception('The provided "minimum_silence_ms" is not a positive number.')

        dBFS = audio.dBFS
        # TODO: Why '- 16' (?) I don't know
        silences = silence.detect_silence(audio, min_silence_len = minimum_silence_ms, silence_thresh = dBFS - 16)

        # [(1.531, 1.946), (..., ...), ...] in seconds
        return [((start / 1000), (stop / 1000)) for start, stop in silences]
    
    @staticmethod
    def create(duration: float, frame_rate: int = 11025, output_filename: Union[str, None] = None):
        """
        Create a silence audio of the given 'duration'. The
        frame rate could be necessary due to different
        videos frame rates.
        
        The file will be stored locally only if
        'output_filename' parameter is provided.
        """
        if not NumberValidator.is_positive_number(duration, do_include_zero = False):
            raise Exception('The provided "duration" is not a positive number.')
        
        # This is the default value for AudioSegment
        frame_rate = 11025 if frame_rate is None else frame_rate

        silence = AudioSegment.silent(duration * 1000, frame_rate)

        """
        if 'output_filename' is True => '.create_temp_filename()'
        if 'output_filename' is str => validate extension and/or fix it with '.create_temp_filename()'
        if 'output_filename' is anything else => NOT STORED
        """

        if output_filename is not None:
            # TODO: Validate output and extension
            silence.export(output_filename, format = 'mp3')

        # TODO: This should be moved to from AudioParser
        # to AudioConverter as it is a conversion
        return AudioParser.as_audioclip(silence)
    
__all__ = [
    'AudioSilence'
]


"""
This is how detected scenes on a video look like when
doing 'print(scene_manager.get_scene_list())':

(00:00:00.000 [frame=0, fps=25.000], 00:00:01.240 [frame=31, fps=25.000])
(00:00:01.240 [frame=31, fps=25.000], 00:00:14.720 [frame=368, fps=25.000])
(00:00:14.720 [frame=368, fps=25.000], 00:00:15.880 [frame=397, fps=25.000])
(00:00:15.880 [frame=397, fps=25.000], 00:00:17.320 [frame=433, fps=25.000])
(00:00:17.320 [frame=433, fps=25.000], 00:00:28.960 [frame=724, fps=25.000])
(00:00:28.960 [frame=724, fps=25.000], 00:00:30.400 [frame=760, fps=25.000])
(00:00:30.400 [frame=760, fps=25.000], 00:01:00.040 [frame=1501, fps=25.000])

As you can see, one scene is a pair of time moments 
and frames, but I think they are actually wrong 
because the end of a scene should be one frame before
the start of the next one.
"""
from yta_multimedia.video.parser import VideoParser
from yta_multimedia.video.frames.video_frame_extractor import VideoFrameExtractor
from yta_multimedia.image.description.blip import describe_image
from yta_general_utils.temp import create_temp_filename
from moviepy.Clip import Clip
from scenedetect import open_video, SceneManager, ContentDetector
from scenedetect.scene_manager import save_images
from typing import Union
from imageio import imsave


# TODO: Turn this into a class and make it possible
# to store in attributes the information and to 
# save images to be able to identify it easy

class VideoScene:
    """
    A class that represent a video scene detected
    by the scene detector.
    """
    video_filename: str = None
    start_time: float = None
    end_time: float = None
    first_frame_number: int = None 
    last_frame_number: int = None
    _first_frame: 'Image' = None
    _last_frame: 'Image' = None

    def __init__(self, video_filename: str, start_time: float, end_time: float, first_frame_number: int, last_frame_number: int):
        # TODO: Make some checkings
        self.video_filename = video_filename
        self.start_time = start_time
        # TODO: I think that the end_time is not okay, it
        # must be 1 frame before because it is actually
        # the next scene first moment, so it is out of this
        # scene. And this is actually failing because it is
        # giving as the end of the last scene of a video 
        # that lasts 1 second exactly (with fps = 25) a 
        # time of 60.04 and a frame of 1501 which is
        # not possible.
        self.end_time = end_time
        self.first_frame_number = first_frame_number
        self.last_frame_number = last_frame_number

    def __str__(self):
        return f'[{str(self.start_time)} (frame = {str(self.first_frame_number)}), {str(self.end_time)} (frame = {str(self.last_frame_number)})]'
    
    @property
    def first_frame(self):
        if self._first_frame is None:
            self._first_frame = VideoFrameExtractor.get_frame_by_time(self.video_filename, self.start_time)

        return self._first_frame
    
    @property
    def last_frame(self):
        if self._last_frame is None:
            self._last_frame = VideoFrameExtractor.get_frame_by_time(self.video_filename, self.end_time)

        return self._last_frame
    
    def describe(self):
        """
        Describe a frame of the scene.
        """
        frame_in_the_middle = int((self.first_frame_number + self.last_frame_number) / 2)
        image = VideoFrameExtractor.get_frame_by_number(self.video_filename, frame_in_the_middle)

        # TODO: The limits are not good images to describe
        return describe_image(image)
    
    def save_first_image(self, output_filename: Union[str, None] = None):
        """
        Save the first image of the scene to the provided
        'output_filename' (or to a temporary file if not
        provided). This method returns the final filename
        that has been used to store the image.
        """
        if not output_filename:
            output_filename = create_temp_filename('tmp_scene_image.png')

        # TODO: Create a better way to write an output
        # that detects the type of object it is receiving
        # and acts consequently
        imsave(output_filename, self.first_frame)

        return output_filename
    
    def save_last_image(self, output_filename: Union[str, None] = None):
        """
        Save the last image of the scene to the provided
        'output_filename' (or to a temporary file if not
        provided). This method returns the final filename
        that has been used to store the image.
        """
        if not output_filename:
            output_filename = create_temp_filename('tmp_scene_image.png')

        # TODO: Create a better way to write an output
        # that detects the type of object it is receiving
        # and acts consequently
        imsave(output_filename, self.last_frame)

        return output_filename

class VideoScenes:
    """
    Class that represent the scenes of a video by using
    a scene detector that detects all the different
    scenes based on a threshold and on consecutive frames
    variation.
    """
    video_filename: str = None
    _scenes: list[VideoScene] = None

    @property
    def scenes(self):
        """
        All the different scenes found in the video.
        """
        if self._scenes is None:
            self._scenes = self._detect_scenes()

        return self._scenes

    def __init__(self, video_filename: str):
        # TODO: Check if video is not filename
        self.video_filename = video_filename

    def _detect_scenes(self):
        video = open_video(self.video_filename, backend = 'moviepy')

        scene_manager = SceneManager()
        scene_manager.add_detector(ContentDetector(threshold = 20))  # 27 is recommended
        scene_manager.detect_scenes(video)

        return [
            VideoScene(
                video_filename = self.video_filename,
                start_time = scene[0].get_seconds(),
                end_time = scene[1].get_seconds(),
                first_frame_number = scene[0].get_frames(),
                last_frame_number = scene[1].get_frames()
            ) for scene in scene_manager.get_scene_list()
            # This is how a scene looks like when printing it:
            # (00:00:00.000 [frame=0, fps=25.000], 00:00:01.240 [frame=31, fps=25.000])
        ]

    # TODO: This should be done by each VideoScene and
    # in another way, manually better, from the original
    # video extracting the grames, I think.
    def save_scenes_images(self, output_dirname: str):
        # TODO: Check that dirname exist
        save_images(
            self.scenes,
            video_manager,
            num_images = 1,
            image_name_template = '$SCENE_NUMBER',
            output_dir = output_dirname
        )

    def __str__(self):
        return '\n'.join([scene.__str__() for scene in self.scenes])


def detect_scenes_from_video(video: Clip):
    """
    Detect the different scenes in the provided 'video'
    (that must be or have a valid filename) based on 
    ambient change (I mean it detects the scene change
    according to color change, etc.).

    This method returns a narray containing the different
    scenes with both 'start' and 'end' fields. Each of
    those fields include a 'second' and 'frame' sub field.

    TODO: This link https://www.scenedetect.com/docs/latest/api.html#example says that
    there is a backend 'VideoStreamMoviePy' class to process it with moviepy. This is
    interesting to use in a clip, and not in a filename.
    """
    # TODO: This import is not working when on top, why (?)
    from scenedetect import SceneManager, ContentDetector

    # This comes from here: https://www.scenedetect.com/
    # Other project: https://github.com/slhck/scenecut-extractor (only ffmpeg)
    video = VideoParser.to_moviepy(video)
    # TODO: I think this has to be a path so it is not able to
    # detect scenes from a video in memory (?)

    # Explained here: https://web.archive.org/web/20160316124732/http://www.bcastell.com/tech-articles/pyscenedetect-tutorial-part-2/

    # TODO: Check if thiss 'filename' includes the abspath
    video = open_video(video.filename, backend = 'moviepy')
    scene_manager = SceneManager()
    scene_manager.add_detector(ContentDetector(threshold = 27))  # 27 is recommended
    scene_manager.detect_scenes(video)

    scenes = []
    for scene in scene_manager.get_scene_list():
        scenes.append({
            'start': {
                'second': scene[0].get_seconds(),
                'frame': scene[0].get_frames()
            },
            'end': {
                'second': scene[1].get_seconds(),
                'frame': scene[1].get_frames()
            }
        })

    return scenes

    # This code below is working but it is quite old
    
    from scenedetect import VideoManager, SceneManager, StatsManager
    from scenedetect.detectors import ContentDetector
    from scenedetect.scene_manager import save_images, write_scene_list_html

    video_manager = VideoManager([video_filename])
    stats_manager = StatsManager()
    scene_manager = SceneManager(stats_manager)

    scene_manager.add_detector(ContentDetector(threshold = 30))
    video_manager.set_downscale_factor()

    video_manager.start()
    scene_manager.detect_scenes(frame_source = video_manager)

    scenes = scene_manager.get_scene_list()
    print(f'{len(scenes)} scenes detected!')

    save_images(
        scenes,
        video_manager,
        num_images = 1,
        image_name_template = '$SCENE_NUMBER',
        output_dir = 'scenes')

    for scene in scenes:
        start, end = scene

        # your code
        print(f'{start.get_seconds()} - {end.get_seconds()}')

    return scenes


# This is to work with abruptness and sooness (https://moviepy.readthedocs.io/en/latest/ref/videofx/moviepy.video.fx.all.accel_decel.html)
    
# I previously had 4.4.2 decorator for moviepy. I forced 4.0.2 and it is apparently working

"""
Interesting:
- https://www.youtube.com/watch?v=Ex1kuxe6jRg (el canal entero tiene buena pinta)
- https://www.youtube.com/watch?v=m6chqKlhpPo Echarle un vistazo a ese tutorial
- https://zulko.github.io/moviepy/ref/videofx.html
- https://stackoverflow.com/questions/48491070/how-to-flip-an-mp4-video-horizontally-in-python
"""



from yta_multimedia.video.parser import VideoParser
from yta_multimedia.video.position import Position, DependantPosition, get_moviepy_upper_left_position
from yta_multimedia.video.edition.duration import set_video_duration
from yta_multimedia.audio.silences import AudioSilence
from yta_multimedia.video.dimensions import resize_video
from yta_general_utils.programming.parameter_validator import PythonValidator
from yta_general_utils.programming.enum import YTAEnum as Enum
from moviepy.Clip import Clip
from moviepy import CompositeVideoClip, CompositeAudioClip
from typing import Union


class VideoCombinatorAudioMode(Enum):
    """
    The mode in which the combined video audios
    will be handled.
    """
    BOTH_CLIPS_AUDIO = 'both_clips_audio'
    """
    Both, the main clip and the added clip audios 
    are preserved.
    """
    ONLY_MAIN_CLIP_AUDIO = 'only_main_clip_audio'
    """
    Only the main clip audio is preserved. The one
    from the added clip is not included.
    """
    ONLY_ADDED_CLIP_AUDIO = 'only_added_clip_audio'
    """
    Only the added clip audio is preserved. The one
    from the main clip is not included.
    """

class VideoCombinator:
    """
    A class to encapsulate and simplify the way we
    combine videos.
    """
    _audio_mode: VideoCombinatorAudioMode = None

    def __init__(self, audio_mode: VideoCombinatorAudioMode = VideoCombinatorAudioMode.BOTH_CLIPS_AUDIO):
        self._audio_mode = VideoCombinatorAudioMode.to_enum(audio_mode) if audio_mode is not None else VideoCombinatorAudioMode.default()

    def video_cover_video(self, video: Clip, background_video: Clip):
        """
        Place the provided 'video' covering the 'background_video'.

        The 'video' will be forced to last the same as the 
        'background_video' to be able to cover it.

        TODO: Explain more
        """
        background_video = VideoParser.to_moviepy(background_video)
        video = VideoParser.to_moviepy(video)

        # We adjust the 'video' size to cover the 
        # 'background_video'
        video = resize_video(video, background_video.size)

        return self.video_over_video(video, background_video, Position.CENTER)

    def video_over_video(self, video: Clip, background_video: Clip, position: Union[Position, DependantPosition, tuple, list]):
        """
        Place the provided 'video' over the 'background_video'
        in the given 'position' (adapted to the real background
        video size). The 'background_video' will be played 
        entirely, and the 'video' clip will be enlarged 
        according to our default enlarging videos strategy. So,
        you should call this method with the parts of the videos
        you want actually combine entirely and pre-processed.

        Pay attention to the size of the videos you provide as
        this method is not considering this part.

        This method returns a CompositeVideoClip with both 
        videos combined.
        """
        if not PythonValidator.is_instance(position, [Position, DependantPosition]) and not PythonValidator.is_tuple(position) and not PythonValidator.is_list(position):
            raise Exception('The provided "position" is not valid, it must be a Position, DependantPosition or a tuple or list with 2 values.')
        
        background_video = VideoParser.to_moviepy(background_video)
        video = VideoParser.to_moviepy(video)

        # TODO: What about sizes (?)

        # TODO: This length strategy is open to changes
        video = set_video_duration(video, background_video.duration)

        # We will place the 'video's center in the 'position'
        # but of the provided 'background_video'
        if PythonValidator.is_instance(position, Position):
            position = position.get_moviepy_upper_left_corner_tuple(video.size, background_video.size)
        elif PythonValidator.is_instance(position, DependantPosition):
            position = position.get_moviepy_position_upper_left_corner(video.size, background_video.size)
        else:
            position = get_moviepy_upper_left_position(background_video.size, video.size, position)

        video = video.with_position(position)

        return CompositeVideoClip([
            background_video,
            video
        ]).with_audio(self._process_audio(background_video, video))

    def _process_audio(self, main_video: Clip, added_video: Clip):
        """
        Process the 'video' and 'background_video' audios
        according to the instance audio mode defined when
        instantiated.

        This method must be called just before combining
        the videos and after video enlargment has been
        applied (if needed).
        """
        # TODO: What about silence 'frame_rate' (?)
        added_video_audio = added_video.audio if added_video.audio is not None else AudioSilence.create(added_video.duration)
        main_video_audio = main_video.audio if main_video.audio is not None else AudioSilence.create(main_video.duration)

        if self._audio_mode == VideoCombinatorAudioMode.BOTH_CLIPS_AUDIO:
            audio = CompositeAudioClip([
                added_video_audio, 
                main_video_audio
            ])
        elif self._audio_mode == VideoCombinatorAudioMode.ONLY_MAIN_CLIP_AUDIO:
            audio = main_video_audio
        elif self._audio_mode == VideoCombinatorAudioMode.ONLY_ADDED_CLIP_AUDIO:
            audio = added_video_audio

        return audio


from yta_multimedia.video.edition.effect.m_effect import MEffect as Effect
from yta_multimedia.video.parser import VideoParser
from yta_multimedia.video.edition.effect.blink_effect import BlinkEffect
from yta_multimedia.video.edition.effect.constants import EFFECTS_RESOURCES_FOLDER
from yta_multimedia.resources.video.effect.sound.drive_urls import PHOTO_GOOGLE_DRIVE_DOWNLOAD_URL
from yta_multimedia.resources import Resource
from moviepy import Clip, AudioFileClip, CompositeAudioClip


class PhotoIsTakenEffect(Effect):
    """
    Simulates that a photo is taken by making a white blink and
    a camera click sound. This effect doesn't freeze the video,
    it is just a white blink with a camera photo sound being
    played.

    This effect doesn't stop or freezes the video so it doesn't
    affect to its duration.

    What to do with the result? **REPLACE**

    The result should be replacing the original clip in the 
    affected part as it is still being reproduced while the
    effect is applied.
    """
    # TODO: Is this working (?)
    result_must_replace = True

    def apply(self, video: Clip) -> Clip:
        # TODO: This is not working properly yet
        #PythonValidator.validate_method_params(BlurEffect.apply, locals(), ['video'])
        video = VideoParser.to_moviepy(video)

        TMP_FILENAME = Resource.get(PHOTO_GOOGLE_DRIVE_DOWNLOAD_URL, EFFECTS_RESOURCES_FOLDER + 'sounds/photo_taken.mp3')

        # We force the effect to be last as much as the clip
        video = BlinkEffect().apply(video, [255, 255, 255])

        # TODO: Maybe raise an Exception as it is too short (?)
        effect_duration = video.duration if video.duration < effect_duration else 0.2

        video.audio = CompositeAudioClip([
            video.audio,
            AudioFileClip(TMP_FILENAME).with_duration(effect_duration)
        ])

        return video



from yta_multimedia.video.parser import VideoParser
from yta_multimedia.video.frames.video_frame_extractor import VideoFrameExtractor
from yta_multimedia.video.generation import generate_video_from_image
from yta_multimedia.video.edition.effect.fit_duration_effect import FitDurationEffect
from yta_general_utils.programming.enum import YTAEnum as Enum
from yta_general_utils.programming.parameter_validator import NumberValidator
from moviepy import VideoClip, concatenate_videoclips
from typing import Union


# TODO: Move this to a better place
class ExtendVideoMode(Enum):
    """
    This is a Enum to set the parameter option to extend the video
    duration with one of these modes (strategies).
    """
    LOOP = 'loop'
    """
    This mode will make the video loop (restart from the begining)
    until it reaches the expected duration.
    """
    FREEZE_LAST_FRAME = 'freeze_last_frame'
    """
    This mode will freeze the last frame of the video and extend 
    it until it reaches the expected duration.
    """
    SLOW_DOWN = 'slow_down'
    """
    This mode will change the speed of the provided video to make
    it fit the needed duration by deccelerating it. As you should
    know, this method changes the whole video duration so the 
    result could be unexpected. Use it carefully.
    """

class EnshortVideoMode(Enum):
    """
    This is a Enum to set the parameter option to enshort
    the video duration with one of these modes (strategies).
    """
    CROP = 'crop'
    """
    This mode will make a subclip of the clip and remove
    the remaining part to fit the expected duration time.
    """
    SPEED_UP = 'speed_up'
    """
    This mode will change the speed of the clip by speeding
    it up. This is useful to make clips shorter when we need
    it, but be careful when you use it. Could be a good 
    choice for transitions that you need to apply in a
    specific amount of time.
    """

def set_video_duration(video: Union[str, VideoClip], duration = float, extend_mode: ExtendVideoMode = ExtendVideoMode.LOOP, enshort_mode: EnshortVideoMode = EnshortVideoMode.CROP):
    """
    This method will return a copy of the provided 'video' with the desired
    'duration' by applying crops, loops or different strategies according
    to the provided 'extend_mode' and 'enshort_mode' parameters. If the
    provided 'duration' is lower than the actual 'video' duration, it will
    be shortened by applying the 'enshort_mode'. If it is greater, it will
    be extended until we reach the desired 'duration' by applying the 
    'extend_mode' provided.

    This method makes a 'video.copy()' internally to work and avoid problems
    with references.
    """
    if not NumberValidator.is_positive_number(duration, do_include_zero = False):
        raise Exception('The provided "duration" parameter is not a positive number.')
    
    video = VideoParser.to_moviepy(video)
    extend_mode = ExtendVideoMode.to_enum(extend_mode) if extend_mode is not None else ExtendVideoMode.default()
    enshort_mode = EnshortVideoMode.to_enum(enshort_mode) if extend_mode is not None else EnshortVideoMode.default()

    final_video = video.copy()

    if video.duration > duration:
        # We need to enshort it
        if enshort_mode == EnshortVideoMode.CROP:
            final_video = final_video.with_subclip(0, duration)
        elif enshort_mode == EnshortVideoMode.SPEED_UP:
            final_video = FitDurationEffect().apply(video, duration)
            #final_video = final_video.with_effects([MultiplySpeed(final_duration = duration)])
            # TODO: Remove this below when effects reviewed and
            # functionality is working perfectly
            #final_video = ChangeSpeedVideoEffect.apply(final_video, duration)
    elif video.duration < duration:
        # We need to enlarge it
        if extend_mode == ExtendVideoMode.LOOP:
            times_to_loop = (int) (duration / video.duration) - 1
            remaining_time = duration % video.duration
            for _ in range(times_to_loop):
                final_video = concatenate_videoclips([final_video, video])
            final_video = concatenate_videoclips([final_video, video.with_subclip(0, remaining_time)])
        elif extend_mode == ExtendVideoMode.FREEZE_LAST_FRAME:
            remaining_time = duration - video.duration
            frame = VideoFrameExtractor.get_frame_by_time(video, video.duration)
            frame_freezed_video = generate_video_from_image(frame, remaining_time)
            final_video = concatenate_videoclips([video, frame_freezed_video])
        elif extend_mode == ExtendVideoMode.SLOW_DOWN:
            final_video = FitDurationEffect().apply(video, duration)
            #final_video = final_video.with_effects([MultiplySpeed(final_duration = duration)])
            # TODO: Remove this below when effects reviewed and
            # functionality is working perfectly
            #final_video = ChangeSpeedVideoEffect.apply(final_video, duration)

    return final_video


from yta_multimedia.video.parser import VideoParser
from moviepy.Clip import Clip
from moviepy import CompositeVideoClip


class VideoEditor:
    _video: Clip = None

    @property
    def video(self):
        return self._video

    def __init__(self, video: Clip):
        self._video = VideoParser.to_moviepy(video, do_include_mask = True, do_calculate_real_duration = True)

    def overlay_text(self, text_generator_wrapping_instance: any):
        # TODO: The instance must be an instance of an
        # specific class (we didn't create yet) to identify
        # it as a manim text animation video generator
        # wrapping class. Maybe identify it as a text that
        # is goin to be overlayed, that can be different
        # from a text that will be the whole scene (imagine
        # a title over a white background vs a text that
        # suddenly appears over what is being shown)
        # 
        # This class contains the needed parameters (with
        # their values actually set) and the animation
        # generator class that must be called with those
        # parameters to generate the animation video.
        
        # If we have the instance here, it is valid
        return CompositeVideoClip([
            self.video,
            VideoParser.to_moviepy(text_generator_wrapping_instance.generate(), do_include_mask = True)
        ])


from yta_multimedia.video.edition.effect.moviepy.position.utils.position import get_moviepy_position
from yta_multimedia.video import MPVideo
from yta_multimedia.video.edition.effect.moviepy.mask import ClipGenerator
from yta_multimedia.video.parser import VideoParser
from yta_general_utils.math.rate_functions import RateFunction
from moviepy.Clip import Clip
from moviepy import CompositeVideoClip
from typing import Union


class MoviepyArgument:
    """
    Arguments used for 'with_position', 'resized' and 'rotated'
    simplified and unified to have the same structure to build
    new videos by modifying the basic moviepy effects.
    """
    def __init__(self, start: Union[int, tuple, float], end: Union[int, tuple, float], t_function: type, rate_func: RateFunction = RateFunction.linear):
        # TODO: 'start' and 'end' must be pure values which
        # means tuples if positions or numbers if resize or
        # rotate
        # TODO: Validate parameters
        self.start = start
        self.end = end
        self.t_function = t_function
        self.rate_func = rate_func
        # TODO: *args or **kwargs should be avoided as we must
        # define the 't_function' with the only parameters 
        # needed and make as much as we need. For example, we
        # should have '.arc_bottom' and '.arc_top' to avoid
        # passing a 'is_bottom' parameter because it would 
        # lost the common structure.

# TODO: Rename to MoviepyWithProcess or something more clear
# TODO: We need to accept a GraphicInterpolation as, at least,
# the 'with_position' parameter and build the needed values
# as it is happening for the example I leave in the 
# video\edition\effect\moviepy\position\move\refactor_this.py
# file code.
class MoviepyWith:
    """
    Class to encapsulate the 'resized', 'with_position' and
    'rotated' functionality and to be able to make one depend
    on the other by pre-calculating the values and then using
    them when positioning, resizing and rotating.

    You should use the same RateFunction if you are planning
    to move and resize the clip at the same time.
    """
    @staticmethod
    def apply(video: Clip, with_position: Union[MoviepyArgument, None] = None, resized: Union[MoviepyArgument, None] = None, rotated: Union[MoviepyArgument, None] = None):
        """
        Apply the property effects on the provided 'video'. This
        method will use a default background video to apply the
        effect.

        Default background is 1920x1080.

        :param MoviepyArgument with_position: The position to apply (as a lambda t function).

        :param MoviepyArgument resized: The resize effect to apply (as a lambda t function).

        :param MoviepyArgument rotated: The rotation effect to apply (as a lambda t function).
        """
        return MoviepyWith.apply_over_video(video, ClipGenerator.get_default_background_video(duration = video.duration), with_position, resized, rotated)

    @staticmethod
    def apply_over_video(video: Clip, background_video: Clip, with_position: Union[MoviepyArgument, None] = None, resized: Union[MoviepyArgument, None] = None, rotated: Union[MoviepyArgument, None] = None):
        """
        Apply the property effects on the provided 'video'. This
        method will use the provided 'background_video' and 
        recalculate the position according to it.

        Background can be different to default scene of 1920x1080
        so positions will be adjusted to it.

        :param MoviepyArgument with_position: The position to apply (as a lambda t function).

        :param MoviepyArgument resized: The resize effect to apply (as a lambda t function).

        :param MoviepyArgument rotated: The rotation effect to apply (as a lambda t function).
        """
        video = VideoParser.to_moviepy(video, do_include_mask = True)
        background_video = VideoParser.to_moviepy(background_video, True)

        # Prepare data for pre-calculations
        video_handler = MPVideo(video)

        background_video = video_handler.prepare_background_clip(background_video)

        if resized:
            resizes = [resized.t_function(t, video_handler.duration, resized.start, resized.end, resized.rate_func) for t in video_handler.frames_time_moments]
        else:
            # We build always this array because position calculation 
            # depends on it to be able to handle movement and resizing
            # at the same time
            resizes = [1 for _ in video_handler.frames_time_moments]

        positions = None
        if with_position:
            # TODO: Maybe this 'get_moviepy_position' can be replaced
            # with the Coordinate scene adapter method
            initial_position = get_moviepy_position(video, background_video, with_position.start)
            final_position = get_moviepy_position(video, background_video, with_position.end)
            positions = []
            for i, t in enumerate(video_handler.frames_time_moments):
                # Calculate each frame position based on each frame size
                initial_position = (initial_position[0] * resizes[i], initial_position[1] * resizes[i])
                final_position = (final_position[0] * resizes[i], final_position[1] * resizes[1])

                # TODO: This 'with_position.t_function' can have different
                # parameters to be able to work, so I don't know how to
                # dynamically handle this. For example, the 'at_position'
                # needs the video.size to be able to calculate where to
                # be centered...
                positions.append(with_position.t_function(t, video.duration, initial_position, final_position))
                
        rotations = None
        if rotated:
            rotations = [rotated.t_function(t, video_handler.duration, rotated.start, rotated.end, rotated.rate_func) for t in video_handler.frames_time_moments]

        # TODO: Maybe avoid the 'resizes' with ones if is the
        # only one we will pass as param

        video = MoviepyWithPrecalculated.apply(video, positions, resizes, rotations)

        return CompositeVideoClip([
            background_video,
            video
        ])

        # I add frame_duration * 0.1 to make sure it fits the
        # next index
        video = video.resized(lambda t: resizes[int((t + frame_duration * 0.1) // frame_duration)])
        if positions:
            video = video.with_position(lambda t: positions[int((t + frame_duration * 0.1) // frame_duration)])
        if rotated:
            video = video.rotated(lambda t: rotations[int((t + frame_duration * 0.1) // frame_duration)])
        
        return video
    
class MoviepyWithPrecalculated:
    """
    Class to encapsulate the 'resized', 'with_position' and
    'rotated' functionality and to be able to make one depend
    on the other by the pre-calculated values provided and
    then using them when positioning, resizing and rotating.

    This class is just to apply the 'with_position', 'resized'
    and 'rotated' moviepy basic effects that must have been
    pre-calculated and passed as a list to be applied for each
    video frame.

    This class is useful when you want to pre-calculate your
    values by a custom function and not only a basic TFuntion
    available through the MoviepyWithArgument class.
    """
    @staticmethod
    def apply(video: Clip, with_position_list: Union[list, None] = None, resized_list: Union[list, None] = None, rotated_list: Union[list, None] = None):
        """
        Apply the property effects on the provided 'video'. This
        method will use the provided 'background_video' and 
        recalculate the position according to it.

        Background can be different to default scene of 1920x1080
        so positions will be adjusted to it.

        :param list with_position: The list of pre-calculated positions to apply.

        :param list resized: The list of pre-calculated resizes to apply.

        :param list rotated: The list of pre-calculated rotation effect to apply.
        """
        return MoviepyWithPrecalculated.apply_over_video(video, ClipGenerator.get_default_background_video(duration = video.duration), with_position_list, resized_list, rotated_list)

    @staticmethod
    def apply_over_video(video: Clip, background_video: Clip, with_position_list: Union[list, None] = None, resized_list: Union[list, None] = None, rotated_list: Union[list, None] = None):
        video = VideoParser.to_moviepy(video, do_include_mask = True)
        background_video = VideoParser.to_moviepy(background_video, True)
        
        # Prepare data for pre-calculations
        video_handler = MPVideo(video)
        
        background_video = video_handler.prepare_background_clip(background_video)

        # Validate provided lists
        if resized_list is not None and len(resized_list) != video_handler.number_of_frames:
            raise Exception(f'The provided "resized_list" contains {str(len(resized_list))} elements and there are "{str(video_handler.number_of_frames)} frames.')
        if with_position_list is not None and len(with_position_list) != video_handler.number_of_frames:
            raise Exception(f'The provided "with_position_list" contains {str(len(with_position_list))} elements and there are "{str(video_handler.number_of_frames)} frames.')
        if rotated_list is not None and len(rotated_list) != video_handler.number_of_frames:
            raise Exception(f'The provided "rotated_list" contains {str(len(rotated_list))} elements and there are "{str(video_handler.number_of_frames)} frames.')
        
        # TODO: Remove this comment below if nothing else 
        # commented below that comment
        # I add frame_duration * 0.1 to make sure it fits the
        # next index
        if resized_list is not None:
            #video = video.resized(lambda t: resized_list[video_handler.get_frame_number_by_time_moment(t)])
            video = video.resized(lambda t: resized_list[video_handler.frame_time_to_frame_index(t, video_handler.fps)])
            #video = video.resized(lambda t: resized_list[int((t + frame_duration * 0.1) // frame_duration)])
        if with_position_list is not None:
            video = video.with_position(lambda t: with_position_list[video_handler.frame_time_to_frame_index(t, video_handler.fps)])
            #video = video.with_position(lambda t: with_position_list[int((t + frame_duration * 0.1) // frame_duration)])
        if rotated_list is not None:
            video = video.with_position(lambda t: rotated_list[video_handler.frame_time_to_frame_index(t, video_handler.fps)])
            #video = video.rotated(lambda t: rotated_list[int((t + frame_duration * 0.1) // frame_duration)])
        
        return CompositeVideoClip([
            background_video,
            video
        ])

